from __future__ import annotations
from typing import Literal, overload
from himena import WidgetDataModel
import impy as ip
from himena.consts import StandardType
from himena.standards.model_meta import (
    ImageMeta,
    ImageChannel,
    ArrayAxis,
)


def image_to_model(
    img: ip.ImgArray | ip.LazyImgArray,
    title: str | None = None,
    is_rgb: bool | None = None,
    orig: WidgetDataModel | None = None,
    is_previewing: bool = False,
) -> WidgetDataModel:
    """Convert impy image array to WidgetDataModel."""
    # normalize is_rgb
    if is_rgb is None and orig is not None:
        meta = orig.metadata
        if isinstance(meta, ImageMeta):
            is_rgb = meta.is_rgb
    if is_rgb is None:
        is_rgb = False

    # normalize channel info
    if "c" in img.axes and not is_rgb:
        channel_axis = img.axes.index("c")
        if channel_labels := img.axes[channel_axis].labels:
            channel_labels = [str(name) for name in channel_labels]
        else:
            channel_labels = [f"Ch-{i}" for i in range(img.shape[channel_axis])]
    else:
        channel_axis = None
        channel_labels = [""]
    nchannels = img.shape[channel_axis] if channel_axis is not None else 1
    meta = ImageMeta(
        axes=[ArrayAxis(name=str(a), scale=a.scale, unit=a.unit) for a in img.axes],
        channel_axis=channel_axis,
        channels=[
            ImageChannel(name=channel_labels[i], contrast_limits=None)
            for i in range(nchannels)
        ],
        is_rgb=is_rgb,
    )
    if orig:
        out = orig.with_value(img.value, title=title, metadata=meta)
        if title is None:
            out = out.with_title_numbering()
    else:
        out = WidgetDataModel(
            value=img.value,
            type=StandardType.IMAGE,
            title=title,
            metadata=meta,
        )
    if is_previewing:
        meta.current_indices = None
        meta.contrast_limits = None
    return out


def label_to_model(
    img: ip.ImgArray | ip.LazyImgArray,
    title: str | None = None,
    is_rgb: bool = False,
    orig: WidgetDataModel | None = None,
    is_previewing: bool = False,
) -> WidgetDataModel:
    out = image_to_model(img, title, is_rgb, orig, is_previewing)
    out.type = StandardType.IMAGE_LABELS
    return out


def make_dims_annotation(model: WidgetDataModel) -> list[tuple[str, int]]:
    if not isinstance(meta := model.metadata, ImageMeta):
        return [("2 (xy)", 2)]
    elif (axes := meta.axes) is None:
        return [("2 (xy)", 2)]
    axis_names = [a.name for a in axes]
    if "z" in axis_names and "y" in axis_names and "x" in axis_names:
        choices = [("2 (xy)", 2), ("3 (xyz)", 3)]
    elif "y" in axis_names and "x" in axis_names:
        choices = [("2 (xy)", 2)]
    elif len(axis_names) > 1:
        if len(axis_names[-1]) == len(axis_names[-2]) == 1:
            label = "".join(axis_names[-2:])
        else:
            label = ", ".join(axis_names[-2:])
        choices = [(f"2 ({label})", 2)]
    else:
        choices = [("2 (xy)", 2)]
    return choices


@overload
def model_to_image(
    model: WidgetDataModel,
    is_previewing: Literal[False] = False,
) -> ip.ImgArray: ...


@overload
def model_to_image(
    model: WidgetDataModel,
    is_previewing: bool = False,
) -> ip.ImgArray | ip.LazyImgArray: ...


def model_to_image(
    model: WidgetDataModel,
    is_previewing: bool = False,
):
    import dask.array as da

    img = model.value
    if not isinstance(meta := model.metadata, ImageMeta):
        raise ValueError("Model must have ImageMeta.")
    if meta.axes is not None:
        scale = {}
        unit = {}
        axes = {}
        for a in meta.axes:
            scale[a.name] = a.scale
            unit[a.name] = a.unit
        axes = list(scale.keys())
    else:
        axes = None
        scale = None
        unit = None
    if isinstance(img, da.Array):
        out = ip.lazy.asarray(img, axes=axes, chunks=img.chunksize)
    elif is_previewing:
        n_multi = img.ndim - 3 if meta.is_rgb else img.ndim - 2
        if n_multi == 0:
            out = ip.asarray(img, axes=axes)
        else:
            chunks_rest = (-1, -1, -1) if meta.is_rgb else (-1, -1)
            out = ip.lazy.asarray(img, axes=axes, chunks=(1,) * n_multi + chunks_rest)
    else:
        out = ip.asarray(img, axes=axes)
    if scale is not None and unit is not None:
        for a in out.axes:
            a.scale = scale.get(str(a))
            a.unit = unit.get(str(a))
    return out



from __future__ import annotations

from functools import singledispatch
import numpy as np
from numpy.typing import NDArray
from scipy import ndimage as ndi

from himena import WidgetDataModel, Parametric, StandardType
from himena.plugins import register_function, configure_gui
from himena.standards import roi, model_meta
from himena.data_wrappers import wrap_array


@singledispatch
def slice_array(r: roi.RoiModel, arr_nd: np.ndarray):
    mask = r.to_mask(arr_nd.shape)
    return arr_nd[..., mask]


@slice_array.register
def _(r: roi.RectangleRoi, arr_nd: np.ndarray):
    bb = r.bbox().adjust_to_int("inner")
    arr = arr_nd[..., bb.top : bb.bottom, bb.left : bb.right]
    return arr.reshape(*arr.shape[:-2], arr.shape[-2] * arr.shape[-1])


@slice_array.register
def _(r: roi.EllipseRoi, arr_nd: np.ndarray):
    bb = r.bbox().adjust_to_int("inner")
    arr = arr_nd[..., bb.top : bb.bottom, bb.left : bb.right]
    _yy, _xx = np.indices(arr.shape[-2:])
    mask = (_yy - r.y) ** 2 / r.height**2 + (_xx - r.x) ** 2 / r.width**2 <= 1
    return arr[..., mask]


@slice_array.register
def _(r: roi.PointRoi2D, arr_nd: np.ndarray):
    out = ndi.map_coordinates(arr_nd, [[r.y], [r.x]], order=1, mode="nearest")
    return out


@slice_array.register
def _(r: roi.PointsRoi2D, arr_nd: np.ndarray):
    coords = np.stack([r.ys, r.xs], axis=1)
    out = ndi.map_coordinates(arr_nd, coords, order=1, mode="nearest")
    return out


@slice_array.register
def _(r: roi.LineRoi, arr_nd: np.ndarray):
    xs, ys = r.arange()
    return _slice_array_along_line(arr_nd, xs, ys)


@slice_array.register
def _(r: roi.SegmentedLineRoi, arr_nd: np.ndarray):
    xs, ys = r.arange()
    return _slice_array_along_line(arr_nd, xs, ys)


def _slice_array_along_line(arr_nd: NDArray[np.number], xs, ys):
    coords = np.stack([ys, xs], axis=1)
    out = np.empty(arr_nd.shape[:-2] + (coords.shape[0],), dtype=np.float32)
    for sl in np.ndindex(arr_nd.shape[:-2]):
        arr_2d = arr_nd[sl]
        out[sl] = ndi.map_coordinates(arr_2d, coords, order=1, mode="nearest")
    return out


@register_function(
    title="Measure ROIs ...",
    types=StandardType.IMAGE,
    menus=["tools/image/roi", "/model_menu/roi"],
    run_async=True,
    command_id="himena-image:roi-measure",
)
def roi_measure(model: WidgetDataModel) -> Parametric:
    metrics_choices = ["mean", "std", "min", "max", "sum"]
    arr = wrap_array(model.value)
    if not isinstance(meta := model.metadata, model_meta.ImageMeta):
        raise ValueError("Image must have an ImageMeta.")
    if axes := meta.axes:
        axis_names = [axis.name for axis in axes[:-2]]
    else:
        axis_names = [f"axis_{i}" for i in range(arr.ndim - 2)]
    axes_choices = [(axis_name, i) for i, axis_name in enumerate(axis_names)]

    @configure_gui(
        metrics={"choices": metrics_choices, "widget_type": "Select"},
        along={"choices": axes_choices, "widget_type": "Select"},
    )
    def run_measure(
        metrics: list[str],
        along: list[int],
    ) -> WidgetDataModel:
        if isinstance(meta.rois, roi.RoiListModel):
            rois = meta.rois
        else:
            rois = meta.rois()
        if len(rois) == 0:
            raise ValueError("No ROIs to measure.")

        ndindex_shape = tuple(arr.shape[i] for i in along)
        funcs = [getattr(np, metric) for metric in metrics]
        out: dict[str, list] = {}
        for along_i in along:
            axis_name = axis_names[along_i]
            out[axis_name] = []
        out["name"] = []
        for metric in metrics:
            if metric in axis_names:
                raise ValueError(
                    f"Name collision between axis names and metrics: {metric!r}"
                )
            out[metric] = []
        for sl in np.ndindex(ndindex_shape):
            for indices, each_roi in rois.iter_with_indices():
                sl_placeholder = list(indices)
                for i, along_i in enumerate(along):
                    sl_placeholder[along_i] = sl[i]
                arr_slice = arr.get_slice(tuple(sl_placeholder))
                target = slice_array(each_roi, arr_slice)
                out["name"].append(each_roi.name)
                for sl_i, axis_name in zip(sl, axis_names):
                    out[axis_name].append(sl_i)
                for func, metric in zip(funcs, metrics):
                    out[metric].append(func(target))
        return WidgetDataModel(
            value=out,
            type=StandardType.DATAFRAME,
            title=f"Results of {model.title}",
        )

    return run_measure



from typing import Literal
from cmap import Colormap

import numpy as np
import impy as ip

from himena import WidgetDataModel, Parametric
from himena.plugins import register_function, configure_gui
from himena.consts import StandardType
from himena.standards.model_meta import ImageMeta, DataFramePlotMeta
from himena.standards import roi
from himena_image.utils import image_to_model, model_to_image

MENU = "image/calculate"


@register_function(
    title="Projection ...",
    menus=MENU,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:projection",
)
def projection(model: WidgetDataModel) -> Parametric:
    """Project the image along an axis."""
    img = model_to_image(model)
    axis_choices = [str(a) for a in img.axes]
    if "z" in axis_choices:
        value = "z"
    elif "t" in axis_choices:
        value = "t"
    else:
        value = axis_choices[0]

    @configure_gui(
        axis={"choices": axis_choices, "value": value, "widget_type": "Select"},
    )
    def run_projection(
        axis: str,
        method: Literal["mean", "max", "min", "sum", "std"],
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.proj(axis=axis, method=method)
        return image_to_model(out, title=model.title)

    return run_projection


@register_function(
    title="Invert",
    menus=MENU,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:invert",
)
def invert(model: WidgetDataModel) -> WidgetDataModel:
    """Invert the image."""
    img = -model.value
    out = model.with_value(img)
    if isinstance(model.metadata, ImageMeta):
        assert isinstance(out.metadata, ImageMeta)
        out.metadata.contrast_limits = None
    return out


@register_function(
    title="Profile line",
    menus=MENU,
    types=[StandardType.IMAGE],
    command_id="himena-image:profile:profile_line",
    keybindings=["/"],
    run_async=True,
)
def profile_line(model: WidgetDataModel) -> Parametric:
    """Get the line profile of the current image slice."""
    if not isinstance(meta := model.metadata, ImageMeta):
        raise ValueError("Metadata is missing.")

    @configure_gui(
        coords={"bind": lambda *_: _get_profile_coords(meta)},
        indices={"bind": lambda *_: _get_indices_channel_composite(meta)},
    )
    def run_profile_line(
        coords: list[list[float]],
        indices: list[int | None],
    ) -> WidgetDataModel:
        img = model_to_image(model)
        _indices = tuple(slice(None) if i is None else i for i in indices)
        img_slice = img[_indices]
        if isinstance(img_slice, ip.LazyImgArray):
            img_slice = img_slice.compute()
        order: int = 0 if img.dtype.kind == "b" else 3
        sliced = img_slice.reslice(coords, order=order)

        if sliced.ndim == 2:  # multi-channel
            sliced_arrays = [sliced[i] for i in range(sliced.shape[0])]
            slice_headers = [
                _channed_name(ch.name, i) for i, ch in enumerate(meta.channels)
            ]
        elif sliced.ndim == 1:
            sliced_arrays = [sliced]
            slice_headers = ["intensity"]
        else:
            raise ValueError(f"Invalid shape: {sliced.shape}.")
        scale = sliced.axes[0].scale
        distance = np.arange(sliced_arrays[0].shape[0]) * scale
        df = {"distance": distance}
        for array, header in zip(sliced_arrays, slice_headers):
            df[header] = array
        color_cycle = [Colormap(ch.colormap or "gray")(0.5).hex for ch in meta.channels]
        return WidgetDataModel(
            value=df,
            type=StandardType.DATAFRAME_PLOT,
            title=f"Profile of {model.title}",
            metadata=DataFramePlotMeta(plot_color_cycle=color_cycle),
        )

    return run_profile_line


@register_function(
    title="Kymograph",
    menus=MENU,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:profile:kymograph",
)
def kymograph(model: WidgetDataModel) -> Parametric:
    """Calculate the kymograph along the specified line."""
    if not isinstance(meta := model.metadata, ImageMeta):
        raise ValueError("Metadata is missing.")

    if meta.current_indices is None:
        raise ValueError("`current_indices` is missing in the image metadata")
    if meta.axes is None:
        raise ValueError("`axes` is missing in the image metadata")
    axes_names = [axis.name for axis in meta.axes]
    if meta.channel_axis is not None:
        axes_names.pop(meta.channel_axis)
    axes_names = axes_names[:-2]  # remove xy

    @configure_gui(
        coords={"bind": _get_profile_coords(meta)},
        along={"choices": axes_names},
    )
    def run_kymograph(
        coords,
        along: str,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        # NOTE: ImgArray supports __getitem__ with dict
        sl: dict[str, int] = {}
        for i, axis in enumerate(img.axes):
            if str(axis) == along:
                continue
            if not hasattr(meta.current_indices[i], "__index__"):
                continue
            sl[str(axis)] = meta.current_indices[i]
        if sl:
            img_slice = img[sl]
        else:
            img_slice = img
        order = 0 if img.dtype.kind == "b" else 3
        sliced = ip.asarray(img_slice.reslice(coords, order=order))
        return image_to_model(sliced, title=f"Kymograph of {model.title}")

    return run_kymograph


def _channed_name(ch: str | None, i: int) -> str:
    if ch is None:
        return f"Ch-{i}"
    return ch


def _get_profile_coords(meta: ImageMeta):
    if isinstance(r := meta.current_roi, roi.LineRoi):
        points = [[r.y1, r.x1], [r.y2, r.x2]]
    elif isinstance(r := meta.current_roi, roi.SegmentedLineRoi):
        points = np.stack([r.ys, r.xs], axis=-1).tolist()
    else:
        raise TypeError(
            "`profile_line` requires a line or segmented line ROI, but the current ROI "
            f"item is {r!r}."
        )
    return points


def _get_indices_channel_composite(meta: ImageMeta):
    """Return the current indices with the channel axis set to None."""
    if meta.current_indices is None:
        raise ValueError("Tried to obtain current indices but it is not set.")
    indices = list(meta.current_indices)
    if meta.channel_axis is not None:
        indices[meta.channel_axis] = None
    indices = tuple(indices)
    return indices



from functools import partial
from pathlib import Path
from typing import Sequence
import zipfile
import impy as ip
import numpy as np
from roifile import ROI_SUBTYPE, ImagejRoi, roiread, roiwrite, ROI_TYPE

from himena import Parametric, StandardType, WidgetDataModel
from himena.consts import MenuId
from himena.standards.model_meta import ImageMeta
from himena.standards import roi as _roi
from himena.plugins import (
    register_reader_plugin,
    register_writer_plugin,
    register_function,
    configure_gui,
)
from himena_image.utils import image_to_model


_SUPPORTED_EXT = frozenset(
    [".tif", ".tiff", ".lsm",
     ".mrc", ".rec", ".st", ".map", ".mrc.gz", ".map.gz",
     ".nd2",
     ]
)  # fmt: skip


@register_reader_plugin
def read_image(path: Path):
    """Read as a image model."""
    img = ip.imread(path)
    is_rgb = "c" in img.axes and path.suffix in [".png", ".jpg", ".jpeg"]
    model = image_to_model(img, is_rgb=is_rgb)
    model.extension_default = path.suffix
    return model


@read_image.define_matcher
def _(path: Path):
    if path.suffix in _SUPPORTED_EXT:
        return StandardType.IMAGE
    return None


@register_writer_plugin
def write_image(model: WidgetDataModel, path: Path):
    """Write image model to a file."""
    img = model.value
    _axes = None
    _scales = {}
    _units = {}
    if isinstance(meta := model.metadata, ImageMeta):
        if axes := meta.axes:
            _axes = [a.name for a in axes]
            _scales = {a.name: a.scale for a in axes}
            _units = {a.name: a.unit for a in axes}
    img = ip.asarray(img, axes=_axes)

    for a in img.axes:
        a.scale = _scales.get(str(a))
        a.unit = _units.get(str(a))
    return img.imsave(path)


@write_image.define_matcher
def _(model: WidgetDataModel, path: Path):
    return model.is_subtype_of(StandardType.ARRAY) and path.suffix in _SUPPORTED_EXT


@register_reader_plugin(priority=-10)
def read_image_as_labels(path: Path):
    """Read as a image model."""
    img = ip.imread(path)
    model = image_to_model(img, is_rgb=False)
    model.extension_default = path.suffix
    model.type = StandardType.IMAGE_LABELS
    return model


@read_image_as_labels.define_matcher
def _(path: Path):
    if path.suffix in _SUPPORTED_EXT:
        return StandardType.IMAGE_LABELS
    return None


@register_reader_plugin
def read_roi(path: Path):
    out = roiread(path)
    if isinstance(out, ImagejRoi):
        ijrois = [out]
    else:
        ijrois = out
    indices: list[Sequence[int]] = []
    rois = []
    for ijroi in ijrois:
        ind, sroi = _to_standard_roi(ijroi)
        indices.append(ind)
        rois.append(sroi)
    axis_names = ["p", "t", "z", "c"]
    indices = np.array(indices, dtype=np.int32)
    val = _roi.RoiListModel(rois, indices=indices, axis_names=axis_names).simplified()
    return WidgetDataModel(value=val, type=StandardType.ROIS, title=path.name)


@read_roi.define_matcher
def _(path: Path):
    ext = "".join(path.suffixes)
    if ext == ".roi":
        return StandardType.ROIS
    elif ext == ".zip":
        with zipfile.ZipFile(path) as z:
            if names := z.namelist():
                if names[0].endswith(".roi"):
                    return StandardType.ROIS
    return None


@register_writer_plugin
def write_roi(model: WidgetDataModel, path: Path):
    if not isinstance(rlist := model.value, _roi.RoiListModel):
        raise ValueError(f"Must be a RoiListModel, got {type(rlist)}")
    _ij_position_getter = partial(
        _to_ij_position, rlist.indices, axis_names=rlist.axis_names
    )
    p_s = _ij_position_getter(["p", "position"])
    t_s = _ij_position_getter(["t", "time"])
    z_s = _ij_position_getter(["z", "slice"])
    c_s = _ij_position_getter(["c", "channel"])
    ijrois: list[ImagejRoi] = []
    for p, t, z, c, roi in zip(p_s, t_s, z_s, c_s, rlist.items):
        multi_dims = {
            "position": p,
            "t_position": t,
            "z_position": z,
            "c_position": c,
        }
        ijrois.append(_from_standard_roi(roi, **multi_dims))
    roiwrite(path, ijrois)
    return None


@write_roi.define_matcher
def _(model: WidgetDataModel, path: Path):
    return model.is_subtype_of(StandardType.ROIS) and path.suffix == ".zip"


@register_function(
    menus=MenuId.FILE,
    title="Open image in lazy mode ...",
    command_id="himena-image:io:lazy-imread",
)
def lazy_imread() -> Parametric:
    @configure_gui
    def run_lazy_imread(path: Path, chunks: list[int]) -> WidgetDataModel:
        img = ip.lazy.imread(path, chunks=chunks)
        model = image_to_model(img)
        model.extension_default = path.suffix
        return model

    return run_lazy_imread


def _get_coords(ijroi: ImagejRoi) -> np.ndarray:
    if ijroi.subpixelresolution:
        return ijroi.subpixel_coordinates - 1
    return ijroi.integer_coordinates - 1 + [ijroi.left, ijroi.top]


def _to_standard_roi(ijroi: ImagejRoi) -> tuple[tuple[int, ...], _roi.RoiModel]:
    p = ijroi.position
    c = ijroi.c_position
    t = ijroi.t_position
    z = ijroi.z_position
    name = ijroi.name

    if ijroi.subtype == ROI_SUBTYPE.UNDEFINED:
        if ijroi.roitype == ROI_TYPE.RECT:
            out = _roi.RectangleRoi(
                x=ijroi.left - 1,
                y=ijroi.top - 1,
                width=ijroi.right - ijroi.left,
                height=ijroi.bottom - ijroi.top,
                name=name,
            )
        elif ijroi.roitype == ROI_TYPE.LINE:
            out = _roi.LineRoi(
                x1=ijroi.x1 - 1,
                y1=ijroi.y1 - 1,
                x2=ijroi.x2 - 1,
                y2=ijroi.y2 - 1,
                name=name,
            )
        elif ijroi.roitype == ROI_TYPE.POINT:
            coords = _get_coords(ijroi)
            if coords.shape[0] == 1:
                out = _roi.PointRoi2D(x=coords[0, 0], y=coords[0, 1], name=name)
            else:
                out = _roi.PointsRoi2D(xs=coords[:, 0], ys=coords[:, 1], name=name)
        elif ijroi.roitype in (ROI_TYPE.POLYGON, ROI_TYPE.FREEHAND):
            coords = _get_coords(ijroi)
            out = _roi.PolygonRoi(xs=coords[:, 0], ys=coords[:, 1], name=name)
        elif ijroi.roitype in (ROI_TYPE.POLYLINE, ROI_TYPE.FREELINE):
            coords = _get_coords(ijroi)
            out = _roi.SegmentedLineRoi(xs=coords[:, 0], ys=coords[:, 1], name=name)
        elif ijroi.roitype == ROI_TYPE.OVAL:
            out = _roi.EllipseRoi(
                x=ijroi.left - 1,
                y=ijroi.top - 1,
                width=ijroi.right - ijroi.left,
                height=ijroi.bottom - ijroi.top,
                name=name,
            )
        else:
            raise ValueError(f"Unsupported ROI type: {ijroi.roitype!r}")
    elif ijroi.subtype == ROI_SUBTYPE.ROTATED_RECT:
        coords = _get_coords(ijroi)
        start = (coords[0] + coords[1]) / 2
        end = (coords[2] + coords[3]) / 2
        width = np.hypot(coords[0, 0] - coords[1, 0], coords[0, 1] - coords[1, 1])
        out = _roi.RotatedRectangleRoi(
            start=tuple(start),
            end=tuple(end),
            width=width,
            name=name,
        )
    elif ijroi.subtype == ROI_SUBTYPE.ELLIPSE:
        # ImageJ rotated ellipse is just a freehand line
        coords = _get_coords(ijroi)
        out = _roi.PolygonRoi(xs=coords[:, 0], ys=coords[:, 1], name=name)
    else:
        raise ValueError(f"Unsupported ROI subtype: {ijroi.subtype}")
    return (p, t, z, c), out


def _from_standard_roi(
    roi: _roi.RoiModel,
    multi_dims: dict[str, int],
) -> ImagejRoi:
    if isinstance(roi, _roi.RectangleRoi):
        return ImagejRoi(
            roitype=ROI_TYPE.RECT,
            name=roi.name,
            top=roi.y + 1,
            left=roi.x + 1,
            bottom=roi.y + roi.height + 1,
            right=roi.x + roi.width + 1,
            **multi_dims,
        )
    elif isinstance(roi, _roi.EllipseRoi):
        return ImagejRoi(
            roitype=ROI_TYPE.OVAL,
            name=roi.name,
            top=roi.y + 1,
            left=roi.x + 1,
            bottom=roi.y + roi.height + 1,
            right=roi.x + roi.width + 1,
            **multi_dims,
        )
    elif isinstance(roi, _roi.LineRoi):
        return ImagejRoi(
            roitype=ROI_TYPE.LINE,
            name=roi.name,
            x1=roi.x1 + 1,
            y1=roi.y1 + 1,
            x2=roi.x2 + 1,
            y2=roi.y2 + 1,
            **multi_dims,
        )
    elif isinstance(roi, _roi.PointRoi2D):
        return ImagejRoi(
            roitype=ROI_TYPE.POINT,
            name=roi.name,
            subpixelresolution=True,
            subpixel_coordinates=np.array([[roi.y + 1, roi.x + 1]]),
            **multi_dims,
        )
    elif isinstance(roi, _roi.PointsRoi2D):
        return ImagejRoi(
            roitype=ROI_TYPE.POINT,
            name=roi.name,
            subpixelresolution=True,
            subpixel_coordinates=np.stack([roi.ys + 1, roi.xs + 1], axis=1),
            **multi_dims,
        )
    elif isinstance(roi, _roi.PolygonRoi):
        return ImagejRoi(
            roitype=ROI_TYPE.POLYGON,
            name=roi.name,
            subpixelresolution=True,
            subpixel_coordinates=np.stack([roi.ys + 1, roi.xs + 1], axis=1),
            **multi_dims,
        )
    elif isinstance(roi, _roi.SegmentedLineRoi):
        return ImagejRoi(
            roitype=ROI_TYPE.POLYLINE,
            name=roi.name,
            subpixelresolution=True,
            subpixel_coordinates=np.stack([roi.ys + 1, roi.xs + 1], axis=1),
            **multi_dims,
        )
    elif isinstance(roi, _roi.RotatedRectangleRoi):
        coords = np.stack(roi._get_vertices(), axis=0) + 1
        return ImagejRoi(
            roitype=ROI_TYPE.FREEHAND,
            subtype=ROI_SUBTYPE.ROTATED_RECT,
            name=roi.name,
            subpixelresolution=True,
            subpixel_coordinates=coords,
            **multi_dims,
        )
    raise ValueError(f"Unsupported ROI type: {type(roi)}")


def _to_ij_position(
    indices: np.ndarray,
    candidates: list[str],
    axis_names: list[str],
) -> np.ndarray:
    for cand in candidates:
        if cand in axis_names:
            return indices[:, axis_names.index(cand)]
    return np.full(indices.shape[0], None, dtype=np.object_)



from himena_image.widgets.viewer import NDImageViewer
from himena.plugins import register_widget_class
from himena.consts import StandardType

register_widget_class(StandardType.IMAGE, NDImageViewer, priority=1)



from typing import Literal, NamedTuple
from himena import WidgetDataModel, Parametric
from himena.plugins import register_function, configure_submenu
from himena_image.utils import image_to_model
import impy as ip

MENU = "file/new/skimage samples"
configure_submenu(MENU, title="scikit-image")


class SampleImage(NamedTuple):
    filename: str
    is_rgb: bool
    axes: str | None = None
    title: str | None = None


SKIMAGE_SAMPLES = [
    SampleImage("astronaut", True, "yxc"),
    SampleImage("brain", False, "zyx"),
    SampleImage("brick", False),
    SampleImage("camera", False),
    SampleImage("cat", True, "yxc"),
    SampleImage("cell", False),
    SampleImage("cells3d", False, "zcyx"),
    SampleImage("checkerboard", False),
    SampleImage("chelsea", True, "yxc"),
    SampleImage("clock", False),
    SampleImage("coffee", True, "yxc"),
    SampleImage("coins", False),
    SampleImage("colorwheel", True, "yxc"),
    SampleImage("eagle", False),
    SampleImage("grass", False),
    SampleImage("gravel", False),
    SampleImage("horse", False),
    SampleImage("hubble_deep_field", True, "yxc"),
    SampleImage("human_mitosis", False),
    SampleImage("immunohistochemistry", True, "yxc"),
    SampleImage("kidney", False, "zyxc"),
    SampleImage("lily", False, "yxc"),
    SampleImage("logo", True, "yxc"),
    SampleImage("microaneurysms", False),
    SampleImage("moon", False),
    SampleImage("nickel_solidification", False, "tyx"),
    SampleImage("page", False),
    SampleImage("palisades_of_vogt", False, "zyx"),
    SampleImage("protein_transport", False, "tcyx"),
    SampleImage("retina", True, "yxc"),
    SampleImage("rocket", True, "yxc"),
    SampleImage("shepp_logan_phantom", False),
    SampleImage("skin", True, "yxc"),
    SampleImage("stereo_motorcycle", True, "yxc"),
    SampleImage("text", False),
    SampleImage("vortex", False),
]


@register_function(title="binary blobs", menus=MENU)
def binary_blobs() -> Parametric:
    from skimage import data

    def make_binary_blobs(
        length: int = 512,
        seed: int = 1,
        blob_size_fraction: float = 0.1,
        volume_fraction: float = 0.5,
        n_dim: Literal[2, 3, 4, 5, 6] = 2,
    ) -> WidgetDataModel[ip.ImgArray]:
        img = data.binary_blobs(
            length=length,
            blob_size_fraction=blob_size_fraction,
            n_dim=n_dim,
            volume_fraction=volume_fraction,
            rng=seed,
        )
        return image_to_model(ip.asarray(img), title="binary blobs")

    return make_binary_blobs


@register_function(title="stereo motorcycle", menus=MENU)
def stereo_motorcycle() -> list[WidgetDataModel[ip.ImgArray]]:
    from skimage import data

    img_left, img_right, disp = data.stereo_motorcycle()
    img_left = ip.asarray(img_left, axes="yxc")
    img_right = ip.asarray(img_right, axes="yxc")
    disp = ip.asarray(disp, axes="yx")
    return [
        image_to_model(img_left, title="Left (stereo_motorcycle)", is_rgb=True),
        image_to_model(img_right, title="Right (stereo_motorcycle)", is_rgb=True),
        image_to_model(disp, title="Disparity (stereo_motorcycle)"),
    ]


def _make_provider(sample: SampleImage):
    def make_sample_image() -> WidgetDataModel[ip.ImgArray]:
        from skimage import data

        img = getattr(data, sample.filename)()
        if sample.title is None:
            title = sample.filename.replace("_", " ").title()
        else:
            title = sample.title
        img = ip.asarray(img, axes=sample.axes)
        if not sample.is_rgb:
            img = img.sort_axes()
        return image_to_model(img, title=title, is_rgb=sample.is_rgb)

    make_sample_image.__name__ = sample.filename
    return make_sample_image


for sample in SKIMAGE_SAMPLES:
    register_function(
        _make_provider(sample),
        title=sample.filename,
        menus=MENU,
        run_async=True,
        command_id=f"himena-image:skimage-sample:{sample}",
    )



from typing import Annotated, Literal

from himena import WidgetDataModel, Parametric
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena.standards.model_meta import ImageMeta
from himena.standards.roi import PointRoi2D
import numpy as np
from himena_image.consts import PaddingMode, InterpolationOrder
from himena_image.utils import make_dims_annotation, image_to_model, model_to_image
from himena_image._mgui_widgets import PointEdit

MENUS = ["image/transform", "/model_menu/transform"]


@register_function(
    title="Shift ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:shift",
    run_async=True,
)
def shift(model: WidgetDataModel) -> Parametric:
    shape = model.value.shape
    if len(shape) < 2:
        raise ValueError("The image must have at least 2 dimensions.")
    max_size = max(shape)

    @configure_gui(
        preview=True,
        shift={
            "options": {
                "widget_type": "FloatSpinBox",
                "min": -max_size,
                "max": max_size,
            }
        },
    )
    def run_shift(
        shift: tuple[float, float],
        mode: PaddingMode = "constant",
        cval: float = 0.0,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.shift(shift, mode=mode, cval=cval, dims=2)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_shift


@register_function(
    title="Rotate ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:rotate",
    run_async=True,
)
def rotate(model: WidgetDataModel) -> Parametric:
    @configure_gui(preview=True)
    def run_rotate(
        degree: Annotated[float, {"min": -90, "max": 90, "widget_type": "FloatSlider"}],
        order: InterpolationOrder = 3,
        mode: PaddingMode = "constant",
        cval: float = 0.0,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        if abs(degree) < 1e-4:
            return model
        img = model_to_image(model, is_previewing)
        out = img.rotate(degree, mode=mode, cval=cval, order=order)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_rotate


@register_function(
    title="Flip ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:flip",
    run_async=True,
)
def flip(model: WidgetDataModel) -> Parametric:
    @configure_gui(preview=True, dimension={"choices": make_dims_annotation(model)})
    def run_flip(
        axis: str,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        idx = img.axes.index(axis)
        slices = [slice(None)] * img.ndim
        slices[idx] = slice(None, None, -1)
        out = img[tuple(slices)]
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_flip


@register_function(
    title="Zoom ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:zoom",
    run_async=True,
)
def zoom(model: WidgetDataModel) -> Parametric:
    @configure_gui(preview=True, dimension={"choices": make_dims_annotation(model)})
    def run_zoom(
        factor: float,
        order: InterpolationOrder = 3,
        mode: PaddingMode = "constant",
        cval: float = 0.0,
        same_shape: bool = False,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.zoom(
            factor,
            order=order,
            mode=mode,
            cval=cval,
            same_shape=same_shape,
            dims=dimension,
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_zoom


@register_function(
    title="Bin ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:bin",
    run_async=True,
)
def bin(model: WidgetDataModel) -> Parametric:
    @configure_gui(preview=True, dimension={"choices": make_dims_annotation(model)})
    def run_bin(
        bin_size: Literal[2, 3, 4, 5, 6, 7, 8],
        method: str = "mean",
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.binning(bin_size, method=method, check_edges=False, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_bin


@register_function(
    title="Radial profile ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:radial_profile",
    run_async=True,
)
def radial_profile(model: WidgetDataModel) -> Parametric:
    def _point_getter(widget):
        if not isinstance(meta := model.metadata, ImageMeta):
            return None
        if not isinstance(roi := meta.current_roi, PointRoi2D):
            return None
        return roi.x, roi.y

    @configure_gui(
        center={"widget_type": PointEdit, "getter": _point_getter},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_radial_profile(
        center: tuple[float, float] | None = None,
        method: Literal["mean", "min", "std", "max"] = "mean",
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.radial_profile(center=center, method=method, dims=dimension)
        return image_to_model(out, orig=model)

    return run_radial_profile


@register_function(
    title="Unmix multi-channel image ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:transform-color:unmix",
)
def unmix(model: WidgetDataModel):
    """Run unmixing of fluorescence leakage between channels."""

    @configure_gui(
        matrix={"types": [StandardType.ARRAY, StandardType.TABLE]},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_unmix(
        matrix: WidgetDataModel,
        background: list[float] | None = None,
    ) -> WidgetDataModel:
        if matrix.type == StandardType.ARRAY:
            matrix = matrix.value
        else:
            matrix = np.asarray(matrix.value).astype(np.float64)
        img = model_to_image(model)
        out = img.unmix(matrix, bg=background)
        return image_to_model(out, orig=model)

    return run_unmix



from __future__ import annotations

from magicgui.widgets.bases import ValuedContainerWidget
from magicgui.widgets import PushButton
from magicgui.types import Undefined
from himena.qt.magicgui import FloatEdit


class PointEdit(ValuedContainerWidget[tuple[float, float]]):
    def __init__(self, point=Undefined, **kwargs):
        kwargs.setdefault("layout", "horizontal")
        getter = kwargs.pop("getter", lambda *_: None)
        if point is Undefined:
            point = (0.0, 0.0)
        self._x_value = FloatEdit(value=point[0], label="x")
        self._y_value = FloatEdit(value=point[1], label="y")
        self._read_btn = PushButton(text="Read")
        self._read_btn.clicked.connect(self._read_point_coords)
        self._getter = getter
        widgets = [self._x_value, self._y_value, self._read_btn]
        super().__init__(value=point, widgets=widgets, **kwargs)
        self.margins = (0, 0, 0, 0)
        self._x_value.changed.connect(self._value_changed)
        self._y_value.changed.connect(self._value_changed)

    def _read_point_coords(self):
        points = self._getter(self)
        if points is not None:
            self.value = points

    def _value_changed(self):
        self.changed.emit(self.value)

    def get_value(self):
        return (self._x_value.value, self._y_value.value)

    def set_value(self, value):
        x, y = value
        self._x_value.value = x
        self._y_value.value = y



from __future__ import annotations

from typing import Any, TYPE_CHECKING, Hashable, Mapping, Sequence
from enum import Enum, auto
from cmap import Colormap
from qtpy import QtWidgets as QtW, QtCore
import numpy as np

from ndv import DataWrapper, ArrayViewer
from superqt import QEnumComboBox
from himena.types import WidgetDataModel
from himena.standards.model_meta import ImageMeta, ArrayAxis
from himena.plugins import validate_protocol

if TYPE_CHECKING:
    from ndv.views._qt._array_view import _QArrayViewer


class ComplexConversionRule(Enum):
    ABS = auto()
    REAL = auto()
    IMAG = auto()
    PHASE = auto()
    LOG_ABS = auto()

    def apply(self, data: np.ndarray) -> np.ndarray:
        if self == ComplexConversionRule.ABS:
            return np.abs(data)
        elif self == ComplexConversionRule.REAL:
            return data.real
        elif self == ComplexConversionRule.IMAG:
            return data.imag
        elif self == ComplexConversionRule.PHASE:
            return np.angle(data)
        elif self == ComplexConversionRule.LOG_ABS:
            return np.log(np.abs(data) + 1e-10)
        raise ValueError(f"Unknown complex conversion rule: {self}")


class ModelDataWrapper(DataWrapper):
    def __init__(self, model: WidgetDataModel):
        super().__init__(model.value)
        if not isinstance(meta := model.metadata, ImageMeta):
            raise ValueError("Invalid metadata")
        self._meta = meta
        self._type = model.type
        self._complex_conversion = ComplexConversionRule.ABS

    @classmethod
    def supports(cls, obj: Any) -> bool:
        return isinstance(obj, WidgetDataModel)

    @property
    def dims(self) -> tuple[str, ...]:
        if axes := self._meta.axes:
            return tuple(a.name for a in axes)
        return tuple(f"axis_{i}" for i in range(len(self._data.shape)))

    @property
    def coords(self) -> Mapping[Hashable, Sequence]:
        """Return the coordinates for the data."""
        return {d: range(s) for d, s in zip(self.dims, self.data.shape)}

    def isel(self, indexers: Mapping[int, int | slice]) -> np.ndarray:
        """Select a slice from a data store using (possibly) named indices."""
        import dask.array as da

        sl = [slice(None)] * len(self._data.shape)
        for k, v in indexers.items():
            sl[k] = v
        out = self._data[tuple(sl)]
        if isinstance(out, da.Array):
            out = out.compute()
        assert isinstance(out, np.ndarray)
        if out.dtype.kind == "b":
            return out.astype(np.uint8)
        elif out.dtype.kind == "c":
            return self._complex_conversion.apply(out)
        return out

    def sizes(self):
        if axes := self._meta.axes:
            names = [a.name for a in axes]
        else:
            names = list(range(len(self._data.shape)))
        return dict(zip(names, self._data.shape))


class NDImageViewer(ArrayViewer):
    def __init__(self):
        super().__init__()
        self._control_widget = QtW.QWidget()
        layout = QtW.QHBoxLayout(self._control_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        spacer = QtW.QWidget()
        spacer.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        layout.addWidget(spacer)
        self._complex_conversion_rule_cbox = QEnumComboBox(
            enum_class=ComplexConversionRule
        )
        self._complex_conversion_rule_cbox.currentEnumChanged.connect(
            self._on_complex_conversion_rule_changed
        )
        layout.addWidget(self._complex_conversion_rule_cbox)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        self.data = model
        is_complex = model.value.dtype.kind == "c"
        self._complex_conversion_rule_cbox.setVisible(is_complex)
        meta = model.metadata
        if isinstance(meta, ImageMeta):
            if meta.is_rgb:
                raise ValueError("RGB images are not supported yet")
            for ch, lut in zip(meta.channels, self.display_model.luts.values()):
                lut.cmap = ch.colormap or "gray"
                lut.clims = ch.contrast_limits
        if is_complex:
            self._complex_conversion_rule_cbox.setCurrentEnum(ComplexConversionRule.ABS)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        indices = [
            None if isinstance(v, slice) else v
            for v in self.display_model.current_index.values()
        ]
        self.display_model.luts
        return WidgetDataModel(
            value=self.data,
            type=self.model_type(),
            metadata=ImageMeta(
                current_indices=indices,
                axes=[ArrayAxis(name=a) for a in self.data_wrapper.dims],
            ),
        )

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return (320, 400)

    @validate_protocol
    def model_type(self) -> str:
        return self.data_wrapper._type

    @validate_protocol
    def native_widget(self) -> _QArrayViewer:
        return self.widget()

    @validate_protocol
    def control_widget(self):
        return self._control_widget

    def _on_complex_conversion_rule_changed(self, enum_: ComplexConversionRule):
        self.data_wrapper._complex_conversion = enum_
        if enum_ is ComplexConversionRule.PHASE:
            cmap_name = "cmocean:phase"
        else:
            cmap_name = "inferno"
        # for ctrl in self._lut_ctrls.values():
        #     ctrl._cmap.setCurrentColormap(cmap.Colormap(cmap_name))
        # self.refresh()
        for val in self.display_model.luts.values():
            val.cmap = Colormap(cmap_name)



from typing import Annotated

from himena import WidgetDataModel, Parametric
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena_image.consts import PaddingMode
from himena_image.utils import make_dims_annotation, image_to_model, model_to_image

MENUS = ["image/morphology", "/model_menu/morphology"]


@register_function(
    title="Dilation ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-morph-basic:dilation",
    run_async=True,
)
def dilation(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_dilation(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.dilation(radius, mode=mode, cval=cval, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_dilation


@register_function(
    title="Erosion ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-morph-basic:erosion",
    run_async=True,
)
def erosion(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_erosion(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.erosion(radius, mode=mode, cval=cval, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_erosion


@register_function(
    title="Opening ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-morph-comp:opening",
    run_async=True,
)
def opening(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_opening(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.opening(radius, mode=mode, cval=cval, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_opening


@register_function(
    title="Closing ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-morph-comp:closing",
    run_async=True,
)
def closing(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_closing(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.closing(radius, mode=mode, cval=cval, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_closing


@register_function(
    title="Top-hat Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-morph-comp:tophat",
    run_async=True,
)
def tophat(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_tophat(
        radius: Annotated[float, {"min": 0.0}] = 30.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.tophat(radius, mode=mode, cval=cval, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_tophat


@register_function(
    title="Skeletonize ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-morph-comp:skeletonize",
    run_async=True,
)
def skeletonize(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_skeletonize(
        radius: Annotated[float, {"min": 0.0}] = 0.0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.skeletonize(radius=radius, dims=dimension)
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_skeletonize
